# Copyright 2014 - Mirantis, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import datetime
import time

from osc_lib.command import command

from mistralclient.commands.v2 import base
from mistralclient import utils


class CronTriggerFormatter(base.MistralFormatter):
    COLUMNS = [
        ('name', 'Name'),
        ('workflow_name', 'Workflow'),
        ('workflow_params', 'Params'),
        ('pattern', 'Pattern'),
        # TODO(rakhmerov): Uncomment when passwords are handled properly.
        # TODO(rakhmerov): Add 'Workflow input' column.
        ('next_execution_time', 'Next execution time'),
        ('remaining_executions', 'Remaining executions'),
        ('created_at', 'Created at'),
        ('updated_at', 'Updated at')
    ]

    @staticmethod
    def format(trigger=None, lister=False):
        if trigger:
            # TODO(rakhmerov): Add following here:
            # TODO(rakhmerov): wf_input = trigger.workflow_input if not lister
            # TODO(rakhmerov:):    else base.cut(trigger.workflow_input)

            data = (
                trigger.name,
                trigger.workflow_name,
                trigger.workflow_params,
                trigger.pattern,
                # TODO(rakhmerov): Uncomment when passwords are handled
                #  properly.
                # TODo(rakhmerov): Add 'wf_input' here.
                trigger.next_execution_time,
                trigger.remaining_executions,
                trigger.created_at,
            )

            if hasattr(trigger, 'updated_at'):
                data += (trigger.updated_at,)
            else:
                data += (None,)
        else:
            data = (tuple('' for _ in
                          range(len(CronTriggerFormatter.COLUMNS))),)

        return CronTriggerFormatter.headings(), data


class List(base.MistralLister):
    """List all cron triggers."""

    def _get_format_function(self):
        return CronTriggerFormatter.format_list

    def _get_resources(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        return mistral_client.cron_triggers.list(
            marker=parsed_args.marker,
            limit=parsed_args.limit,
            sort_keys=parsed_args.sort_keys,
            sort_dirs=parsed_args.sort_dirs,
            fields=CronTriggerFormatter.fields(),
            **base.get_filters(parsed_args)
        )


class Get(command.ShowOne):
    """Show specific cron trigger."""

    def get_parser(self, prog_name):
        parser = super(Get, self).get_parser(prog_name)

        parser.add_argument('cron_trigger', help='Cron trigger name')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        return CronTriggerFormatter.format(mistral_client.cron_triggers.get(
            parsed_args.cron_trigger
        ))


class Create(command.ShowOne):
    """Create new trigger."""

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument('name', help='Cron trigger name')
        parser.add_argument('workflow_identifier', help='Workflow name or ID')

        parser.add_argument(
            'workflow_input',
            nargs='?',
            help='Workflow input'
        )

        parser.add_argument(
            '--params',
            help='Workflow params',
        )

        parser.add_argument(
            '--pattern',
            type=str,
            help='Cron trigger pattern',
            metavar='<* * * * *>'
        )
        parser.add_argument(
            '--first-time',
            type=str,
            default=None,
            help=("Date and time of the first execution. Time is treated as "
                  "local time unless --utc is also specified"),
            metavar='<YYYY-MM-DD HH:MM>'
        )
        parser.add_argument(
            '--count',
            type=int,
            help="Number of wanted executions",
            metavar='<integer>'
        )
        parser.add_argument(
            '--utc',
            action='store_true',
            help="All times specified should be treated as UTC"
        )

        return parser

    @staticmethod
    def _get_file_content_or_dict(string):
        if string:
            return utils.load_json(string)
        else:
            return {}

    @staticmethod
    def _convert_time_string_to_utc(time_string):
        datetime_format = '%Y-%m-%d %H:%M'

        the_time = time_string
        if the_time:
            the_time = datetime.datetime.strptime(
                the_time, datetime_format)

            is_dst = time.daylight and time.localtime().tm_isdst > 0
            utc_offset = - (time.altzone if is_dst else time.timezone)

            the_time = (the_time - datetime.timedelta(
                0, utc_offset)).strftime(datetime_format)

        return the_time

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        wf_input = self._get_file_content_or_dict(parsed_args.workflow_input)
        wf_params = self._get_file_content_or_dict(parsed_args.params)

        first_time = parsed_args.first_time
        if not parsed_args.utc:
            first_time = self._convert_time_string_to_utc(
                parsed_args.first_time)

        trigger = mistral_client.cron_triggers.create(
            parsed_args.name,
            parsed_args.workflow_identifier,
            wf_input,
            wf_params,
            parsed_args.pattern,
            first_time,
            parsed_args.count
        )

        return CronTriggerFormatter.format(trigger)


class Delete(command.Command):
    """Delete trigger."""

    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'cron_trigger',
            nargs='+', help='Name of cron trigger(s).'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        utils.do_action_on_many(
            lambda s: mistral_client.cron_triggers.delete(s),
            parsed_args.cron_trigger,
            "Request to delete cron trigger %s has been accepted.",
            "Unable to delete the specified cron trigger(s)."
        )



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy

from oslo_utils import importutils

from mistralclient.api import httpclient
from mistralclient.api.v2 import action_executions
from mistralclient.api.v2 import actions
from mistralclient.api.v2 import code_sources
from mistralclient.api.v2 import cron_triggers
from mistralclient.api.v2 import dynamic_actions
from mistralclient.api.v2 import environments
from mistralclient.api.v2 import event_triggers
from mistralclient.api.v2 import executions
from mistralclient.api.v2 import members
from mistralclient.api.v2 import services
from mistralclient.api.v2 import tasks
from mistralclient.api.v2 import workbooks
from mistralclient.api.v2 import workflows
from mistralclient import auth


osprofiler_profiler = importutils.try_import("osprofiler.profiler")

_DEFAULT_MISTRAL_URL = "http://localhost:8989/v2"


class Client(object):

    MANAGERS = {
        'workbooks': workbooks.WorkbookManager,
        'executions': executions.ExecutionManager,
        'tasks': tasks.TaskManager,
        'actions': actions.ActionManager,
        'workflows': workflows.WorkflowManager,
        'cron_triggers': cron_triggers.CronTriggerManager,
        'event_triggers': event_triggers.EventTriggerManager,
        'environments': environments.EnvironmentManager,
        'action_executions': action_executions.ActionExecutionManager,
        'services': services.ServiceManager,
        'members': members.MemberManager,
        'code_sources': code_sources.CodeSourceManager,
        'dynamic_actions': dynamic_actions.DynamicActionManager,
        }

    def __init__(self, auth_type='keystone', **kwargs):
        # We get the session at this point, as some instances of session
        # objects might have mutexes that can't be deep-copied.
        session = kwargs.pop('session', None)
        enforce_raw_definitions = kwargs.pop('enforce_raw_definitions', False)
        req = copy.deepcopy(kwargs)
        mistral_url = req.get('mistral_url')
        profile = req.get('profile')

        if mistral_url and not isinstance(mistral_url, str):
            raise RuntimeError('Mistral url should be a string.')

        # If auth url was provided then we perform an authentication, otherwise
        # just ignore this step
        if req.get('auth_url') or req.get('target_auth_url'):
            auth_handler = auth.get_auth_handler(auth_type)
            auth_response = auth_handler.authenticate(req, session=session)
        else:
            auth_response = {}

        if session is None:
            # If the session was None and we're using keystone auth, it will be
            # created by the auth_handler.
            session = auth_response.pop('session', None)

        req.update(auth_response)

        mistral_url = auth_response.get('mistral_url') or mistral_url

        if not mistral_url:
            mistral_url = _DEFAULT_MISTRAL_URL

        if profile:
            osprofiler_profiler.init(profile)

        http_client = httpclient.HTTPClient(mistral_url, session=session,
                                            **req)
        # Create all resource managers.
        for name, cls in self.MANAGERS.items():
            setattr(self, name, cls(http_client, enforce_raw_definitions))



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""OpenStackClient plugin for Workflow service."""

import logging

from osc_lib import utils

LOG = logging.getLogger(__name__)

DEFAULT_WORKFLOW_API_VERSION = '2'
API_VERSION_OPTION = 'os_workflow_api_version'
API_NAME = 'workflow_engine'
API_VERSIONS = {
    '2': 'mistralclient.api.v2.client.Client',
}


def make_client(instance):
    """Returns a workflow_engine service client."""
    version = instance._api_version[API_NAME]
    workflow_client = utils.get_client_class(
        API_NAME,
        version,
        API_VERSIONS)

    LOG.debug('Instantiating workflow engine client: %s', workflow_client)

    mistral_url = instance.get_endpoint_for_service_type(
        'workflowv2',
        region_name=instance.region_name,
        interface='publicURL'
    )

    client = workflow_client(mistral_url=mistral_url, session=instance.session)

    return client


def build_option_parser(parser):
    """Hook to add global options."""
    parser.add_argument(
        '--os-workflow-api-version',
        metavar='<workflow-api-version>',
        default=utils.env(
            'OS_WORKFLOW_API_VERSION',
            default=DEFAULT_WORKFLOW_API_VERSION),
        help='Workflow API version, default=' +
             DEFAULT_WORKFLOW_API_VERSION +
             ' (Env: OS_WORKFLOW_API_VERSION)')

    return parser



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistralclient.api.v2 import action_executions
from mistralclient.tests.unit.v2 import base

# TODO(everyone): later we need additional tests verifying all the errors etc.

ACTION_EXEC = {
    'id': "1",
    'name': 'my_action_execution',
    'workflow_name': 'my_wf',
    'state': 'RUNNING',
}

URL_TEMPLATE = '/action_executions'
URL_TEMPLATE_ID = '/action_executions/%s'


class TestActionExecutions(base.BaseClientV2Test):
    def test_create(self):
        self.requests_mock.post(self.TEST_URL + URL_TEMPLATE,
                                json=ACTION_EXEC,
                                status_code=201)

        body = {
            'name': ACTION_EXEC['name']
        }

        action_execution = self.action_executions.create(
            'my_action_execution',
            {}
        )

        self.assertIsNotNone(action_execution)
        self.assertEqual(action_executions.ActionExecution(
            self.action_executions, ACTION_EXEC
        ).to_dict(), action_execution.to_dict())

        self.assertEqual(body, self.requests_mock.last_request.json())

    def test_update(self):
        url = self.TEST_URL + URL_TEMPLATE_ID % ACTION_EXEC['id']
        self.requests_mock.put(url, json=ACTION_EXEC)

        body = {
            'state': ACTION_EXEC['state']
        }

        action_execution = self.action_executions.update(
            ACTION_EXEC['id'],
            ACTION_EXEC['state']
        )

        self.assertIsNotNone(action_execution)

        expected = action_executions.ActionExecution(
            self.action_executions,
            ACTION_EXEC
        ).to_dict()

        self.assertEqual(
            expected,
            action_execution.to_dict()
        )

        self.assertEqual(body, self.requests_mock.last_request.json())

    def test_list(self):
        self.requests_mock.get(self.TEST_URL + URL_TEMPLATE,
                               json={'action_executions': [ACTION_EXEC]})

        action_execution_list = self.action_executions.list()

        self.assertEqual(1, len(action_execution_list))
        action_execution = action_execution_list[0]

        expected = action_executions.ActionExecution(
            self.action_executions,
            ACTION_EXEC
        ).to_dict()

        self.assertEqual(expected, action_execution.to_dict())

    def test_list_with_limit(self):
        self.requests_mock.get(
            self.TEST_URL + URL_TEMPLATE,
            json={'action_executions': [ACTION_EXEC]}
        )

        action_execution_list = self.action_executions.list(limit=1)

        self.assertEqual(1, len(action_execution_list))

        last_request = self.requests_mock.last_request

        # Make sure that limit is passed to the server correctly.
        self.assertEqual(['1'], last_request.qs['limit'])

    def test_list_with_no_limit(self):
        self.requests_mock.get(
            self.TEST_URL + URL_TEMPLATE,
            json={'action_executions': [ACTION_EXEC]}
        )

        action_execution_list = self.action_executions.list(limit=-1)

        self.assertEqual(1, len(action_execution_list))

        last_request = self.requests_mock.last_request

        self.assertNotIn('limit', last_request.qs)

    def test_get(self):
        url = self.TEST_URL + URL_TEMPLATE_ID % ACTION_EXEC['id']
        self.requests_mock.get(url, json=ACTION_EXEC)

        action_execution = self.action_executions.get(ACTION_EXEC['id'])

        expected = action_executions.ActionExecution(
            self.action_executions,
            ACTION_EXEC
        ).to_dict()

        self.assertEqual(expected, action_execution.to_dict())

    def test_delete(self):
        url = self.TEST_URL + URL_TEMPLATE_ID % ACTION_EXEC['id']
        self.requests_mock.delete(url, status_code=204)

        self.action_executions.delete(ACTION_EXEC['id'])



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse

from cliff import command
from cliff import show

from mistralclient.commands.v2 import base
from mistralclient import utils


class WorkflowFormatter(base.MistralFormatter):
    COLUMNS = [
        ('id', 'ID'),
        ('name', 'Name'),
        ('namespace', 'Namespace'),
        ('project_id', 'Project ID'),
        ('tags', 'Tags'),
        ('input', 'Input'),
        ('scope', 'Scope'),
        ('created_at', 'Created at'),
        ('updated_at', 'Updated at')
    ]

    @staticmethod
    def format(workflow=None, lister=False):
        if workflow:
            tags = getattr(workflow, 'tags', None) or []

            data = (
                workflow.id,
                workflow.name,
                workflow.namespace,
                workflow.project_id,
                base.wrap(', '.join(tags)) or '<none>',
                workflow.input if not lister else base.cut(workflow.input),
                workflow.scope,
                workflow.created_at
            )

            if hasattr(workflow, 'updated_at'):
                data += (workflow.updated_at,)
            else:
                data += (None,)
        else:
            data = (tuple('' for _ in range(len(WorkflowFormatter.COLUMNS))),)

        return WorkflowFormatter.headings(), data


class List(base.MistralLister):
    """List all workflows."""

    def _get_format_function(self):
        return WorkflowFormatter.format_list

    def get_parser(self, prog_name):
        parser = super(List, self).get_parser(prog_name)

        return parser

    def _get_resources(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        return mistral_client.workflows.list(
            marker=parsed_args.marker,
            limit=parsed_args.limit,
            sort_keys=parsed_args.sort_keys,
            sort_dirs=parsed_args.sort_dirs,
            fields=WorkflowFormatter.fields(),
            **base.get_filters(parsed_args)
        )


class Get(show.ShowOne):
    """Show specific workflow."""

    def get_parser(self, prog_name):
        parser = super(Get, self).get_parser(prog_name)

        parser.add_argument('workflow', help='Workflow ID or name.')
        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Namespace to get the workflow from.",
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        wf = mistral_client.workflows.get(
            parsed_args.workflow,
            parsed_args.namespace
        )

        return WorkflowFormatter.format(wf)


class Create(base.MistralLister):
    """Create new workflow."""

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            'definition',
            type=argparse.FileType('r'),
            help='Workflow definition file.'
        )
        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Namespace to create the workflow within.",
        )
        parser.add_argument(
            '--public',
            action='store_true',
            help='With this flag workflow will be marked as "public".'
        )

        return parser

    def _get_format_function(self):
        return WorkflowFormatter.format_list

    def _validate_parsed_args(self, parsed_args):
        if not parsed_args.definition:
            raise RuntimeError("You must provide path to workflow "
                               "definition file.")

    def _get_resources(self, parsed_args):
        scope = 'public' if parsed_args.public else 'private'

        mistral_client = self.app.client_manager.workflow_engine

        return mistral_client.workflows.create(
            parsed_args.definition.read(),
            namespace=parsed_args.namespace,
            scope=scope
        )


class Delete(command.Command):
    """Delete workflow."""

    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'workflow',
            nargs='+',
            help='Name or ID of workflow(s).'
        )

        parser.add_argument(
            '--namespace',
            nargs='?',
            default=None,
            help="Namespace to delete the workflow from.",
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        utils.do_action_on_many(
            lambda s: mistral_client.workflows.delete(s,
                                                      parsed_args.namespace),
            parsed_args.workflow,
            "Request to delete workflow %s has been accepted.",
            "Unable to delete the specified workflow(s)."
        )


class Update(base.MistralLister):
    """Update workflow."""

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument(
            'definition',
            type=argparse.FileType('r'),
            help='Workflow definition'
        )

        parser.add_argument('--id', help='Workflow ID.')

        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Namespace of the workflow.",
        )

        parser.add_argument(
            '--public',
            action='store_true',
            help='With this flag workflow will be marked as "public".'
        )

        return parser

    def _get_format_function(self):
        return WorkflowFormatter.format_list

    def _get_resources(self, parsed_args):
        scope = 'public' if parsed_args.public else 'private'

        mistral_client = self.app.client_manager.workflow_engine

        return mistral_client.workflows.update(
            parsed_args.definition.read(),
            scope=scope,
            id=parsed_args.id,
            namespace=parsed_args.namespace
        )


class GetDefinition(command.Command):
    """Show workflow definition."""

    def get_parser(self, prog_name):
        parser = super(GetDefinition, self).get_parser(prog_name)

        parser.add_argument('identifier', help='Workflow ID or name.')
        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Namespace to get the workflow from.",
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        wf = mistral_client.workflows.get(
            parsed_args.identifier,
            parsed_args.namespace
        )

        self.app.stdout.write(wf.definition or "\n")


class Validate(show.ShowOne):
    """Validate workflow."""
    @staticmethod
    def _format(result=None):
        columns = ('Valid', 'Error')

        if result:
            data = (result.get('valid'), result.get('error'),)
        else:
            data = (tuple('<none>' for _ in range(len(columns))),)

        return columns, data

    def get_parser(self, prog_name):
        parser = super(Validate, self).get_parser(prog_name)

        parser.add_argument(
            'definition',
            type=argparse.FileType('r'),
            help='Workflow definition file'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        result = mistral_client.workflows.validate(
            parsed_args.definition.read()
        )

        return self._format(result)



# Copyright 2014 Mirantis, Inc.
# Copyright 2020 Nokia Software.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock


from mistralclient.api.v2 import actions
from mistralclient.commands.v2 import actions as action_cmd
from mistralclient.commands.v2 import base as cmd_base
from mistralclient.tests.unit import base


ACTION_DICT = {
    'id': '1234-4567-7894-7895',
    'name': 'a',
    'is_system': True,
    'input': "param1",
    'description': 'My cool action',
    'tags': ['test'],
    'created_at': '1',
    'updated_at': '1',
    'namespace': 'test_namespace'
}

ACTION_DEF = """
---
version: '2.0'

base: std.echo
base-parameters:
    output: "<% $.str1 %><% $.str2 %>"
output: "<% $ %><% $ %>"
"""

ACTION_WITH_DEF_DICT = ACTION_DICT.copy()
ACTION_WITH_DEF_DICT.update({'definition': ACTION_DEF})
ACTION = actions.Action(mock, ACTION_DICT)
ACTION_WITH_DEF = actions.Action(mock, ACTION_WITH_DEF_DICT)


class TestCLIActionsV2(base.BaseCommandTest):
    @mock.patch('argparse.open', create=True)
    def test_create(self, mock_open):
        self.client.actions.create.return_value = [ACTION]

        result = self.call(action_cmd.Create, app_args=['1.txt'])

        self.assertEqual(
            [('1234-4567-7894-7895', 'a', True, "param1",
              'My cool action', 'test', '1', '1')],
            result[1]
        )

    @mock.patch('argparse.open', create=True)
    def test_create_public(self, mock_open):
        self.client.actions.create.return_value = [ACTION]

        result = self.call(
            action_cmd.Create,
            app_args=['1.txt', '--public']
        )

        self.assertEqual(
            [('1234-4567-7894-7895', 'a', True, "param1",
              'My cool action', 'test', '1', '1')],
            result[1]
        )

        self.assertEqual(
            'public',
            self.client.actions.create.call_args[1]['scope']
        )

    @mock.patch('argparse.open', create=True)
    def test_create_long_input(self, mock_open):
        action_long_input_dict = ACTION_DICT.copy()
        long_input = ', '.join(
            ['var%s' % i for i in range(10)]
        )
        action_long_input_dict['input'] = long_input
        workflow_long_input = actions.Action(
            mock.Mock(),
            action_long_input_dict
        )
        self.client.actions.create.return_value = [workflow_long_input]

        result = self.call(action_cmd.Create, app_args=['1.txt'])

        self.assertEqual(
            [('1234-4567-7894-7895', 'a',
              True, cmd_base.cut(long_input), 'My cool action',
              'test', '1', '1')],
            result[1]
        )

    @mock.patch('argparse.open', create=True)
    def test_update(self, mock_open):
        self.client.actions.update.return_value = [ACTION]

        result = self.call(action_cmd.Update, app_args=['my_action.yaml'])

        self.assertEqual(
            [('1234-4567-7894-7895', 'a', True, "param1",
              'My cool action', 'test', '1', '1')],
            result[1]
        )

    @mock.patch('argparse.open', create=True)
    def test_update_public(self, mock_open):
        self.client.actions.update.return_value = [ACTION]

        result = self.call(
            action_cmd.Update,
            app_args=['my_action.yaml', '--public']
        )

        self.assertEqual(
            [('1234-4567-7894-7895', 'a', True, "param1",
              'My cool action', 'test', '1', '1')],
            result[1]
        )

        self.assertEqual(
            'public',
            self.client.actions.update.call_args[1]['scope']
        )

    def test_list(self):
        self.client.actions.list.return_value = [ACTION]

        result = self.call(action_cmd.List)

        self.assertEqual(
            [('1234-4567-7894-7895', 'a', True, "param1",
              'My cool action', 'test', '1', '1')],
            result[1]
        )

    def test_get(self):
        self.client.actions.get.return_value = ACTION

        result = self.call(action_cmd.Get, app_args=['name'])

        self.assertEqual(
            ('1234-4567-7894-7895', 'a', True, "param1",
             'My cool action', 'test', '1', '1'),
            result[1]
        )

    def test_delete(self):
        self.call(action_cmd.Delete, app_args=['name'])

        self.client.actions.delete.assert_called_once_with('name',
                                                           namespace='')

    def test_delete_with_namespace(self):
        self.call(action_cmd.Delete, app_args=['name',
                                               '--namespace',
                                               'test_namespace']
                  )

        self.client.actions.delete.assert_called_once_with(
            'name',
            namespace='test_namespace')

    def test_delete_with_multi_names_and_namespace(self):
        self.call(action_cmd.Delete, app_args=['name1',
                                               'name2',
                                               '--namespace',
                                               'test_namespace'])

        self.assertEqual(2, self.client.actions.delete.call_count)
        self.assertEqual(
            [mock.call('name1', namespace='test_namespace'),
             mock.call('name2', namespace='test_namespace')],
            self.client.actions.delete.call_args_list
        )

    def test_delete_with_multi_names(self):
        self.call(action_cmd.Delete, app_args=['name1', 'name2'])

        self.assertEqual(2, self.client.actions.delete.call_count)
        self.assertEqual(
            [mock.call('name1', namespace=''),
             mock.call('name2', namespace='')],
            self.client.actions.delete.call_args_list
        )

    def test_get_definition(self):
        self.client.actions.get.return_value = ACTION_WITH_DEF

        self.call(action_cmd.GetDefinition, app_args=['name'])

        self.app.stdout.write.assert_called_with(ACTION_DEF)

    @mock.patch('argparse.open', create=True)
    def test_validate(self, mock_open):
        self.client.actions.validate.return_value = {'valid': True}

        result = self.call(action_cmd.Validate, app_args=['action.yaml'])

        self.assertEqual((True, None), result[1])

    @mock.patch('argparse.open', create=True)
    def test_validate_failed(self, mock_open):
        self.client.actions.validate.return_value = {
            'valid': False,
            'error': 'Invalid DSL...'
        }

        result = self.call(action_cmd.Validate, app_args=['action.yaml'])

        self.assertEqual((False, 'Invalid DSL...'), result[1])



# Copyright (c) 2014 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import os_client_config
from tempest.lib.cli import base


CLI_DIR = os.environ.get(
    'OS_MISTRALCLIENT_EXEC_DIR',
    os.path.join(os.path.abspath('.'), '.tox/functional/bin')
)


def credentials(cloud='devstack-admin'):
    """Retrieves credentials to run functional tests

    Credentials are either read via os-client-config from the environment
    or from a config file ('clouds.yaml'). Environment variables override
    those from the config file.
    devstack produces a clouds.yaml with two named clouds - one named
    'devstack' which has user privs and one named 'devstack-admin' which
    has admin privs. This function will default to getting the devstack-admin
    cloud as that is the current expected behavior.
    """
    return get_cloud_config(cloud=cloud).get_auth_args()


def get_cloud_config(cloud='devstack-admin'):
    return os_client_config.OpenStackConfig().get_one_cloud(cloud=cloud)


class MistralCLIAuth(base.ClientTestBase):

    _mistral_url = None

    def _get_admin_clients(self):
        creds = credentials()

        clients = base.CLIClient(
            username=creds['username'],
            password=creds['password'],
            tenant_name=creds['project_name'],
            project_name=creds['project_name'],
            user_domain_id=creds['user_domain_id'],
            project_domain_id=creds['project_domain_id'],
            uri=creds['auth_url'],
            cli_dir=CLI_DIR
        )

        return clients

    def _get_clients(self):
        return self._get_admin_clients()

    def mistral(self, action, flags='', params='', fail_ok=False):
        """Executes Mistral command."""
        mistral_url_op = "--os-mistral-url %s" % self._mistral_url
        flags = "{} --insecure".format(flags)

        if 'WITHOUT_AUTH' in os.environ:
            return base.execute(
                'mistral %s' % mistral_url_op,
                action,
                flags,
                params,
                fail_ok,
                merge_stderr=False,
                cli_dir=''
            )
        else:
            return self.clients.cmd_with_auth(
                'mistral %s' % mistral_url_op,
                action,
                flags,
                params,
                fail_ok
            )

    def get_project_id(self, project_name='admin'):
        admin_clients = self._get_clients()

        projects = self.parser.listing(
            admin_clients.openstack(
                'project show',
                params=project_name,
                flags='--os-identity-api-version 3 --insecure'
            )
        )

        return [o['Value'] for o in projects if o['Field'] == 'id'][0]


class MistralCLIAltAuth(base.ClientTestBase):

    _mistral_url = None

    def _get_alt_clients(self):
        creds = credentials('devstack-alt')

        clients = base.CLIClient(
            username=creds['username'],
            password=creds['password'],
            project_name=creds['project_name'],
            tenant_name=creds['project_name'],
            user_domain_id=creds['user_domain_id'],
            project_domain_id=creds['project_domain_id'],
            uri=creds['auth_url'],
            cli_dir=CLI_DIR
        )

        return clients

    def _get_clients(self):
        return self._get_alt_clients()

    def mistral_alt(self, action, flags='', params='', mode='alt_user'):
        """Executes Mistral command for alt_user from alt_tenant."""
        mistral_url_op = "--os-mistral-url %s" % self._mistral_url
        flags = "{} --insecure".format(flags)

        return self.clients.cmd_with_auth(
            'mistral %s' % mistral_url_op, action, flags, params)



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import logging
import os.path

from oslo_serialization import jsonutils

from osc_lib.command import command

from cliff.lister import Lister as cliff_lister
from mistralclient.commands.v2 import base
from mistralclient import utils

LOG = logging.getLogger(__name__)


class ExecutionFormatter(base.MistralFormatter):
    COLUMNS = [
        ('id', 'ID'),
        ('workflow_id', 'Workflow ID'),
        ('workflow_name', 'Workflow name'),
        ('workflow_namespace', 'Workflow namespace'),
        ('description', 'Description'),
        ('task_execution_id', 'Task Execution ID'),
        ('root_execution_id', 'Root Execution ID'),
        ('state', 'State'),
        ('state_info', 'State info'),
        ('created_at', 'Created at'),
        ('updated_at', 'Updated at'),
        ('duration', 'Duration', True),
    ]

    @staticmethod
    def format(wf_ex=None, lister=False):
        if wf_ex:
            state_info = (
                wf_ex.state_info if not lister
                else base.cut(wf_ex.state_info)
            )

            duration = base.get_duration_str(
                wf_ex.created_at,
                wf_ex.updated_at if wf_ex.state in ['ERROR', 'SUCCESS'] else ''
            )

            data = (
                wf_ex.id,
                wf_ex.workflow_id,
                wf_ex.workflow_name,
                wf_ex.workflow_namespace,
                wf_ex.description,
                wf_ex.task_execution_id or '<none>',
                wf_ex.root_execution_id or '<none>',
                wf_ex.state,
                state_info,
                wf_ex.created_at,
                wf_ex.updated_at or '<none>',
                duration
            )
        else:
            data = (tuple('' for _ in
                          range(len(ExecutionFormatter.COLUMNS))),)

        return ExecutionFormatter.headings(), data


class List(base.MistralExecutionLister):
    """List all executions."""

    def _get_format_function(self):
        return ExecutionFormatter.format_list

    def get_parser(self, parsed_args):
        parser = super(List, self).get_parser(parsed_args)

        parser.add_argument(
            '--task',
            nargs='?',
            help="Parent task execution ID associated with workflow "
                 "execution list.",
        )
        parser.add_argument(
            '--rootsonly',
            help="return only root executions",
            action='store_true',
        )

        return parser

    def _get_resources(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        none_fields = 'root_execution_id' if parsed_args.rootsonly else ''

        return mistral_client.executions.list(
            nulls=none_fields,
            task=parsed_args.task,
            marker=parsed_args.marker,
            limit=parsed_args.limit,
            sort_keys=parsed_args.sort_keys,
            sort_dirs=parsed_args.sort_dirs,
            fields=ExecutionFormatter.fields(),
            **base.get_filters(parsed_args)
        )


class Get(command.ShowOne):
    """Show specific execution."""

    def get_parser(self, prog_name):
        parser = super(Get, self).get_parser(prog_name)

        parser.add_argument('execution', help='Execution identifier')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        execution = mistral_client.executions.get(parsed_args.execution)

        return ExecutionFormatter.format(execution)


class Create(command.ShowOne):
    """Create new execution."""

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            'workflow_identifier',
            nargs='?',
            help='Workflow ID or name'
        )
        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Workflow namespace."
        )
        parser.add_argument(
            'workflow_input',
            nargs='?',
            help='Workflow input'
        )
        parser.add_argument(
            'params',
            nargs='?',
            help='Workflow additional parameters'
        )
        parser.add_argument(
            '-d',
            '--description',
            dest='description',
            default='',
            help='Execution description'
        )
        parser.add_argument(
            '-s',
            dest='source_execution_id',
            nargs='?',
            help="Workflow Execution id which will allow operators to create "
                 "a new workflow execution based on the previously successful "
                 "executed workflow. Example: mistral execution-create -s "
                 "123e4567-e89b-12d3-a456-426655440000")

        return parser

    def take_action(self, parsed_args):
        if parsed_args.workflow_input:
            wf_input = utils.load_json(parsed_args.workflow_input)
        else:
            wf_input = {}

        if parsed_args.params:
            params = utils.load_json(parsed_args.params)
        else:
            params = {}

        mistral_client = self.app.client_manager.workflow_engine

        execution = mistral_client.executions.create(
            parsed_args.workflow_identifier,
            parsed_args.namespace,
            wf_input,
            parsed_args.description,
            parsed_args.source_execution_id,
            **params
        )

        return ExecutionFormatter.format(execution)


class Delete(command.Command):
    """Delete execution."""

    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'execution',
            nargs='+',
            help='Id of execution identifier(s).'
        )

        parser.add_argument(
            '--force',
            default=False,
            action='store_true',
            help='Force the deletion of an execution. Might cause a cascade '
                 ' of errors if used for running executions.'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        force = parsed_args.force

        utils.do_action_on_many(
            lambda s: mistral_client.executions.delete(s, force=force),
            parsed_args.execution,
            "Request to delete execution %s has been accepted.",
            "Unable to delete the specified execution(s)."
        )


class Update(command.ShowOne):
    """Update execution."""

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument('id', help='Execution identifier')

        parser.add_argument(
            '-s',
            '--state',
            dest='state',
            choices=['RUNNING', 'PAUSED', 'SUCCESS', 'ERROR', 'CANCELLED'],
            help='Execution state'
        )

        parser.add_argument(
            '-e',
            '--env',
            dest='env',
            help='Environment variables'
        )

        parser.add_argument(
            '-d',
            '--description',
            dest='description',
            help='Execution description'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        env = (
            utils.load_file(parsed_args.env)
            if parsed_args.env and os.path.isfile(parsed_args.env)
            else utils.load_content(parsed_args.env)
        )

        execution = mistral_client.executions.update(
            parsed_args.id,
            parsed_args.state,
            description=parsed_args.description,
            env=env
        )

        return ExecutionFormatter.format(execution)


class GetInput(command.Command):
    """Show execution input data."""

    def get_parser(self, prog_name):
        parser = super(GetInput, self).get_parser(prog_name)

        parser.add_argument('id', help='Execution ID')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        ex_input = mistral_client.executions.get(parsed_args.id).input

        try:
            ex_input = jsonutils.loads(ex_input)
            ex_input = jsonutils.dumps(ex_input, indent=4) + "\n"
        except Exception:
            LOG.debug("Execution input is not JSON.")

        self.app.stdout.write(ex_input or "\n")


class GetOutput(command.Command):
    """Show execution output data."""

    def get_parser(self, prog_name):
        parser = super(GetOutput, self).get_parser(prog_name)

        parser.add_argument('id', help='Execution ID')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        output = mistral_client.executions.get(parsed_args.id).output

        try:
            output = jsonutils.loads(output)
            output = jsonutils.dumps(output, indent=4) + "\n"
        except Exception:
            LOG.debug("Execution output is not JSON.")

        self.app.stdout.write(output or "\n")


REPORT_ENTRY_INDENT = 4


class GetReport(command.Command):
    """Print execution report."""

    def get_parser(self, prog_name):
        parser = super(GetReport, self).get_parser(prog_name)

        parser.add_argument('id', help='Execution ID')
        parser.add_argument(
            '--errors-only',
            dest='errors_only',
            action='store_true',
            help='Only error paths will be included.'
        )
        parser.add_argument(
            '--statistics-only',
            dest='statistics_only',
            action='store_true',
            help='Only the statistics will be included.'
        )
        parser.add_argument(
            '--no-errors-only',
            dest='errors_only',
            action='store_false',
            help='Not only error paths will be included.'
        )
        parser.set_defaults(errors_only=True)

        parser.add_argument(
            '--max-depth',
            dest='max_depth',
            nargs='?',
            type=int,
            default=-1,
            help='Maximum depth of the workflow execution tree. '
                 'If 0, only the root workflow execution and its '
                 'tasks will be included'
        )

        return parser

    def print_line(self, line, level=0):
        self.app.stdout.write(
            "%s%s\n" % (' ' * (level * REPORT_ENTRY_INDENT), line)
        )

    def print_workflow_execution_entry(self, wf_ex, level):
        self.print_line(
            "workflow '%s' [%s] %s" %
            (wf_ex['name'], wf_ex['state'], wf_ex['id']),
            level
        )

        if 'task_executions' in wf_ex:
            for t_ex in wf_ex['task_executions']:
                self.print_task_execution_entry(t_ex, level + 1)

    def print_task_execution_entry(self, t_ex, level):
        self.print_line(
            "task '%s' [%s] %s" %
            (t_ex['name'], t_ex['state'], t_ex['id']),
            level
        )

        if 'retry_count' in t_ex:
            self.print_line('(retry count: %s)' % t_ex['retry_count'], level)

        if t_ex['state'] == 'ERROR':
            state_info = t_ex['state_info']

            if state_info:
                state_info = state_info[0:100].replace('\n', ' ') + '...'

                self.print_line('(error info: %s)' % state_info, level)

        if 'action_executions' in t_ex:
            for a_ex in t_ex['action_executions']:
                self.print_action_execution_entry(a_ex, level + 1)

        if 'workflow_executions' in t_ex:
            for wf_ex in t_ex['workflow_executions']:
                self.print_workflow_execution_entry(wf_ex, level + 1)

    def print_action_execution_entry(self, a_ex, level):
        self.print_line(
            "action '%s' [%s] %s" %
            (a_ex['name'], a_ex['state'], a_ex['id']),
            level
        )

        if a_ex['state'] == 'ERROR':
            state_info = a_ex['state_info']

            if state_info:
                state_info = state_info[0:100] + '...'

                self.print_line('(error info: %s)' % state_info, level)

    def print_statistics(self, stat):
        self.print_line(
            'Number of tasks in SUCCESS state: %s' %
            stat['success_tasks_count']
        )
        self.print_line(
            'Number of tasks in ERROR state: %s' % stat['error_tasks_count']
        )
        self.print_line(
            'Number of tasks in RUNNING state: %s' %
            stat['running_tasks_count']
        )
        self.print_line(
            'Number of tasks in IDLE state: %s' % stat['idle_tasks_count']
        )
        self.print_line(
            'Number of tasks in PAUSED state: %s\n' %
            stat['paused_tasks_count']
        )

        if 'estimated_time' in stat:
            self.print_line(
                'Estimated time (seconds) for the execution to finish:'' %s\n'
                % stat['estimated_time']
            )

    def print_report(self, report_json):
        self.print_line(
            "\nTo get more details on a task failure "
            "run: mistral task-get <id> -c 'State info'\n"
        )

        frame_line = '=' * 30

        self.print_line(
            '%s General Statistics %s\n' %
            (frame_line, frame_line)
        )
        self.print_statistics(report_json['statistics'])

        if 'root_workflow_execution' in report_json:
            self.print_line(
                '%s Workflow Execution Tree %s\n' %
                (frame_line, frame_line)
            )
            self.print_workflow_execution_entry(
                report_json['root_workflow_execution'],
                0
            )

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        report_json = mistral_client.executions.get_report(
            parsed_args.id,
            errors_only=parsed_args.errors_only,
            max_depth=parsed_args.max_depth,
            statistics_only=parsed_args.statistics_only,
        )

        self.print_report(report_json)


class GetPublished(command.Command):
    """Show workflow global published variables."""

    def get_parser(self, prog_name):
        parser = super(GetPublished, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            help='Workflow ID')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        res = mistral_client.executions.get(parsed_args.id)
        published = res.published_global \
            if hasattr(res, 'published_global') else None

        try:
            published = jsonutils.loads(published)
            published = jsonutils.dumps(published, indent=4) + "\n"
        except Exception:
            LOG.debug("Task result is not JSON.")

        self.app.stdout.write(published or "\n")


class SubExecutionsBaseLister(cliff_lister):

    def _get_format_function(self):
        raise NotImplementedError

    def _get_resources_function(self):
        raise NotImplementedError

    def get_parser(self, prog_name):
        parser = super(SubExecutionsBaseLister, self).get_parser(prog_name)

        parser.add_argument(
            'id',
            metavar='ID',
            help='origin id'
        )
        parser.add_argument(
            '--errors-only',
            dest='errors_only',
            action='store_true',
            help='Only error paths will be included.'
        )
        parser.add_argument(
            '--max-depth',
            dest='max_depth',
            nargs='?',
            type=int,
            default=-1,
            help='Maximum depth of the workflow execution tree. '
                 'If 0, only the root workflow execution and its '
                 'tasks will be included'
        )

        return parser

    def _get_resources(self, parsed_args):
        resource_function = self._get_resources_function()
        errors_only = parsed_args.errors_only or ''

        return resource_function(
            parsed_args.id,
            errors_only=errors_only,
            max_depth=parsed_args.max_depth,
        )

    def take_action(self, parsed_args):
        format_func = self._get_format_function()
        execs_list = self._get_resources(parsed_args)

        if not isinstance(execs_list, list):
            execs_list = [execs_list]

        data = [format_func(r)[1] for r in execs_list]

        return (format_func()[0], data) if data else format_func()


class SubExecutionsLister(SubExecutionsBaseLister):

    def _get_format_function(self):
        return ExecutionFormatter.format_list

    def _get_resources_function(self):
        mistral_client = self.app.client_manager.workflow_engine

        return mistral_client.executions.get_ex_sub_executions
